import java.util.Hashtable;
import java.util.Scanner;
import java.util.Vector;
import java.io.*;
import org.antlr.runtime.*;
import org.antlr.runtime.tree.*;

public class KasusBC {
	CommonTree root;
	CodeBlock result;

	Hashtable<String, Integer> variables = new Hashtable<String, Integer>();
	Hashtable<String, Instruction> binary_inst_map;

	Vector<String> strings = new Vector<String>();

	KasusBC(CommonTree root) {
		this.root = root;		
		binary_inst_map = new Hashtable<String, Instruction>();
		binary_inst_map.put("<",Instruction.CMPLT);		
		binary_inst_map.put(">",Instruction.CMPGT);		
		binary_inst_map.put(">=",Instruction.CMPGE);		
		binary_inst_map.put("<=",Instruction.CMPLE);		
		binary_inst_map.put("==",Instruction.CMPEQ);		
		binary_inst_map.put("!=",Instruction.CMPNE);		
		binary_inst_map.put("+",Instruction.ADD);		
		binary_inst_map.put("-",Instruction.SUB);		
		binary_inst_map.put("*",Instruction.MUL);		
		binary_inst_map.put("%",Instruction.MOD);		
		binary_inst_map.put("/",Instruction.DIV);		
	}

	void checkVar(String var) throws Exception {
		if (!variables.containsKey(var)) {
			throw new Exception("Variable '"+var+"' not declared");
		}
	}

	CodeBlock evaluateExpr(CommonTree expr) throws Exception{
		CommonTree left = (CommonTree)expr.getChild(0);
		CommonTree  right = null;
		if (expr.getChildCount()==2) {
			right = (CommonTree)expr.getChild(1);
		}
		if (expr.getType()==KasusLexer.INT) {
			int v = Integer.parseInt(expr.getText());
			return new CodeBlock(Instruction.PUSH, v);
		}
		if (expr.getType()==KasusLexer.ID) {
			checkVar(expr.getText());
			return new CodeBlock(Instruction.PUSHVAR,
					     (int)variables.get(expr.getText()));
		}
		if (expr.getText().equals("+")) {
			if (expr.getChildCount()==1) {
				return evaluateExpr(left);
			}
		}
		if (expr.getText().equals("-")) {
			if (expr.getChildCount()==1) {
				CodeBlock e = evaluateExpr(left);
				e.addInstruction(Instruction.NEG, 0);
				return e;
			}
		}
		CodeBlock b1 = evaluateExpr(left);
		CodeBlock b2 = evaluateExpr(right);
		b1.append(b2);
		b1.addInstruction(binary_inst_map.get(expr.getText()), 0);
		return b1;
	}

	CodeBlock evaluateDeclaration(CommonTree  decl) {
		int ctr = 0;
		for (Object e: decl.getChildren()) {
			CommonTree et = (CommonTree)e;
			variables.put(et.getText(), ctr++);
		}
		return new CodeBlock(); //empty
	}

	CodeBlock evaluateInput(CommonTree expr) throws Exception {
		String var = expr.getText();
		checkVar(var);
		CodeBlock result = new CodeBlock();
		result.addInstruction(Instruction.CALL,
				      Instruction.INPUT);
		result.addInstruction(Instruction.POP,
				      variables.get(var));
		return result;
	}

	CodeBlock evaluatePrint(int type, CommonTree expr) throws Exception {
		CodeBlock result = new CodeBlock();
		if (expr.getType()==KasusLexer.STRING) {
			int stringnumber = strings.size();
			String s = expr.getText();
			/*remove first ' and last ' from string*/
			s = s.replaceAll("^'", "");
			s = s.replaceAll("'$", "");
			strings.add(s);
			result.addInstruction(Instruction.PUSHSTR,
					      stringnumber);
			result.addInstruction(Instruction.CALL,
					      Instruction.PRINTSTR);

		} else {
			String var = expr.getText();
			checkVar(var);
			result.addInstruction(Instruction.PUSHVAR,
					      variables.get(var));
			result.addInstruction(Instruction.CALL,
					      Instruction.PRINTVAR);
		}
		if (type==KasusLexer.PRINTLN) {
			result.addInstruction(Instruction.CALL,
					      Instruction.PRINTLINE);
		}
		return result;
	}

	CodeBlock evaluateAssignment(CommonTree astmt) throws Exception {
		CommonTree var = (CommonTree)astmt.getChild(0);
		checkVar(var.getText());
		CommonTree expr = (CommonTree)astmt.getChild(1);
		CodeBlock result = new CodeBlock();
		result.append(evaluateExpr(expr));
		result.addInstruction(Instruction.POP,
				      variables.get(var.getText()));	   
		return result;
	}

	CodeBlock evaluateIf(CommonTree astmt) throws Exception {
		CommonTree expr = (CommonTree)astmt.getChild(0);
		CommonTree ifpart = (CommonTree)astmt.getChild(1);

		CodeBlock res = evaluateExpr(expr);
		CodeBlock ifblock = evaluateStatementList(ifpart);
		
		if (astmt.getChildCount()==2) { /*if*/
			res.addInstruction(Instruction.JUMPFALSE,
					   ifblock.getLength()+1);
			res.append(ifblock);
		} else { /*if else*/
			CommonTree elsepart = (CommonTree)astmt.getChild(2);
			CodeBlock elseblock = evaluateStatementList(elsepart);
			ifblock.addInstruction(Instruction.JUMP, 
					       elseblock.getLength()+1);			
			res.addInstruction(Instruction.JUMPFALSE,
					   ifblock.getLength()+1);
			res.append(ifblock);
			res.append(elseblock);
		}
		return res;
	}

	CodeBlock evaluateWhile(CommonTree astmt) throws Exception {
		CommonTree expr = (CommonTree)astmt.getChild(0);
		CommonTree whilepart = (CommonTree)astmt.getChild(1);
		CodeBlock result = evaluateExpr(expr);
		CodeBlock statements =  evaluateStatementList(whilepart);

		result.addInstruction(Instruction.JUMPFALSE,
					   statements.getLength()+2);

		statements.addInstruction(Instruction.JUMP,
					  -(statements.getLength() + 
					   result.getLength()));
		result.append(statements);
		return result;
	}

	CodeBlock evaluateStatement(CommonTree expr) throws Exception{
		switch (expr.getType()) {
		case KasusLexer.DECLARATION:
			return evaluateDeclaration(expr);
		case KasusLexer.INPUT:
			return evaluateInput((CommonTree)expr.getChild(0));
		case KasusLexer.PRINT:
		case KasusLexer.PRINTLN:
			return evaluatePrint(expr.getType(), (CommonTree)expr.getChild(0));
		case KasusLexer.ASSIGNMENT:
			return evaluateAssignment(expr);
		case KasusLexer.IF_STATEMENT:
			return evaluateIf(expr);
		case KasusLexer.WHILE_STATEMENT:
			return evaluateWhile(expr);
		}
		return null; //should not happen
	}

	CodeBlock evaluateStatementList(CommonTree exprlist) throws Exception{
		CodeBlock cb = new CodeBlock();
		for (Object e: exprlist.getChildren()) {
			CodeBlock st = evaluateStatement((CommonTree)e);
			cb.append(st);
		}
		return cb;
	}

	void evaluate() throws Exception {
		result =  evaluateStatementList(root);
	}

	void writeToFile(String filename) throws IOException {
		FileOutputStream fos = new FileOutputStream(filename);
		DataOutputStream dos = new DataOutputStream(fos);
		/*write number of variables*/
		dos.writeInt(variables.size());
		/*write number of strings*/
		dos.writeInt(strings.size());
		/*write all the strings*/
		for (String str: strings) {
			dos.writeUTF(str);
		}		
		result.write(dos);
		fos.close();
	}

	public static void main(String argv[]) throws Exception {
		if (argv.length!=1) {
			System.out.println("Usage: KasusBC <filename.e>");
			return;
		}
		KasusLexer lex = new KasusLexer(new ANTLRFileStream(argv[0]));
		CommonTokenStream tokens = new CommonTokenStream(lex);
		KasusParser g = new KasusParser(tokens);
		KasusParser.prog_return r = g.prog();
		CommonTree ct = (CommonTree)r.getTree();
		KasusBC el = new KasusBC(ct);
		el.evaluate();
		el.writeToFile(argv[0]+".bc");
		
	}
}
